from functools import cached_property
from pathlib import Path
from typing import Optional


class Env:
    _instance = None

    # The working directory path
    wspath: Optional[Path] = None

    # The current task path
    taskpath: Optional[Path] = None

    # Set to True when multi-processing when
    # in slave mode:
    # - no progress report
    slave: bool = False

    @cached_property
    def xpm_path(self):
        path = self.taskpath / ".experimaestro"
        path.mkdir(exist_ok=True)
        return path

    @staticmethod
    def instance():
        if Env._instance is None:
            Env._instance = Env()
        return Env._instance



# Configuration registers

from typing import ClassVar, Dict, Optional, Set, Type, Union

from pathlib import Path
import typing
from omegaconf import DictConfig, OmegaConf, SCMode
import pkg_resources
from experimaestro.utils import logger

from .base import ConnectorConfiguration, TokenConfiguration
from .specs import HostRequirement

if typing.TYPE_CHECKING:
    from experimaestro.launchers import Launcher
    from experimaestro.tokens import Token


class LauncherNotFoundError(Exception):
    pass


Connectors = Dict[str, Dict[str, ConnectorConfiguration]]
Tokens = Dict[str, Dict[str, TokenConfiguration]]


def load_yaml(schema, path: Path):
    if not path.is_file():
        return {}

    logger.debug("Loading %s", path)
    with path.open("rt") as fp:
        cfg = OmegaConf.load(fp)
        return OmegaConf.to_container(
            OmegaConf.merge(cfg, schema), structured_config_mode=SCMode.INSTANTIATE
        )


class LauncherRegistry:
    INSTANCES: ClassVar[Dict[Path, "LauncherRegistry"]] = {}
    CURRENT_CONFIG_DIR: ClassVar[Optional[Path]] = None

    @staticmethod
    def instance():
        """Returns an instance for the current configuration directory"""
        if LauncherRegistry.CURRENT_CONFIG_DIR is None:
            LauncherRegistry.CURRENT_CONFIG_DIR = Path(
                "~/.config/experimaestro"
            ).expanduser()

        if LauncherRegistry.CURRENT_CONFIG_DIR not in LauncherRegistry.INSTANCES:
            LauncherRegistry.INSTANCES[
                LauncherRegistry.CURRENT_CONFIG_DIR
            ] = LauncherRegistry(LauncherRegistry.CURRENT_CONFIG_DIR)

        return LauncherRegistry.INSTANCES[LauncherRegistry.CURRENT_CONFIG_DIR]

    @staticmethod
    def set_config_dir(config_dir: Path):
        LauncherRegistry.CURRENT_CONFIG_DIR = config_dir

    def __init__(self, basepath: Path):
        self.connectors_schema = DictConfig({})
        self.tokens_schema = DictConfig({})
        self.find_launcher_fn = None

        # Use entry points for connectors and launchers
        for entry_point in pkg_resources.iter_entry_points("experimaestro.connectors"):
            entry_point.load().init_registry(self)

        for entry_point in pkg_resources.iter_entry_points("experimaestro.tokens"):
            entry_point.load().init_registry(self)

        # Register the find launcher function if it exists
        launchers_py = basepath / "launchers.py"
        if launchers_py.is_file():
            logger.info("Loading %s", launchers_py)

            from importlib import util

            with launchers_py.__fspath__() as fp:
                spec = util.spec_from_file_location("xpm_launchers_conf", fp)
                module = util.module_from_spec(spec)
                spec.loader.exec_module(module)

            self.find_launcher_fn = getattr(module, "find_launcher", None)
            if self.find_launcher_fn is None:
                logger.warning(
                    "No find_launcher() function was found in %s", launchers_py
                )

        # Read the configuration file
        self.connectors = load_yaml(
            self.connectors_schema, basepath / "connectors.yaml"
        )
        self.tokens = load_yaml(self.tokens_schema, basepath / "tokens.yaml")

    def register_connector(self, identifier: str, cls: Type):
        self.connectors_schema.merge_with({identifier: cls})

    def register_token(self, identifier: str, cls: Type):
        self.tokens_schema.merge_with({identifier: cls})

    def getToken(self, identifier: str) -> "Token":
        for tokens in self.tokens.values():
            if identifier in tokens:
                return tokens[identifier].create(self, identifier)
        raise AssertionError(f"No token with identifier {identifier}")

    def getConnector(self, identifier: str):
        for connectors in self.connectors.values():
            if identifier in connectors:
                return connectors[identifier].create(self)

        # Default local connector
        if identifier == "local":
            from experimaestro.connectors.local import LocalConnector

            return LocalConnector.instance()

        raise AssertionError(f"No connector with identifier {identifier}")

    def find(
        self, *input_specs: Union[HostRequirement, str], tags: Set[str] = set()
    ) -> Optional["Launcher"]:
        """ "
        Arguments:
            spec: The processing requirements
            tags: Restrict the launchers to those containing one of the specified tags
        """

        if self.find_launcher_fn is None:
            logger.info("No launchers.yaml file: using local host ")
            from experimaestro.launchers.direct import DirectLauncher
            from experimaestro.connectors.local import LocalConnector

            return DirectLauncher(LocalConnector.instance())

        # Parse specs
        from .parser import parse

        specs = []
        for spec in input_specs:
            if isinstance(spec, str):
                specs.extend(parse(spec))
            else:
                specs.append(spec)

        # Use launcher function
        if self.find_launcher_fn is not None:
            for spec in specs:
                if launcher := self.find_launcher_fn(spec, tags):
                    return launcher

        return None


def find_launcher(
    *specs: Union[HostRequirement, str], tags: Set[str] = set()
) -> "Launcher":
    """Find a launcher matching a given specification"""
    launcher = LauncherRegistry.instance().find(*specs, tags=tags)
    if not launcher:
        raise LauncherNotFoundError(
            f"No launcher with specification: {specs}."
            "Please refer to the documentation at the following URL: "
            "https://experimaestro-python.readthedocs.io/en/latest/launchers/"
        )
    return launcher



"""Handles dynamic task outputs"""

import asyncio
import json
import logging
import queue
import threading
from collections import defaultdict
from functools import cached_property
from pathlib import Path
from typing import Callable, TYPE_CHECKING

from watchdog.events import FileSystemEventHandler

from experimaestro.ipc import ipcom
from experimaestro.utils import logger

from .base import Job, experiment

if TYPE_CHECKING:
    from experimaestro.core.objects import WatchedOutput


class TaskOutputCallbackHandler:
    def __init__(self, converter: Callable):
        pass


class TaskOutputs(FileSystemEventHandler):
    """Represent and monitors dynamic outputs generated by one task"""

    #: Global dictionary for handles
    HANDLERS: dict[Path, "TaskOutputs"] = {}

    #: Global lock to access current HANDLERS
    LOCK = threading.Lock()

    def create(job: Job):
        with TaskOutputs.LOCK:
            if instance := TaskOutputs.get(job.task_outputs_path, None):
                return instance

            instance = TaskOutputs(job.task_outputs_path)
            TaskOutputs[job.task_outputs_path] = instance
            return instance

    def __init__(self, path: Path):
        """Monitors an event path"""
        logger.debug("Watching dynamic task outputs in %s", path)
        self.path = path
        self.handle = None
        self.count = 0
        self.lock = threading.Lock()
        self.listeners: dict[str, dict[Callable, set[Callable]]] = defaultdict(
            lambda: defaultdict(set)
        )

        #: The events registered so far
        self.events = []

    def __enter__(self):
        """Starts monitoring task outputs"""
        self.job.task_outputs_path.parent.mkdir(parents=True, exist_ok=True)
        with self.lock:
            if self.handle is None:
                assert self.count == 0
                self.handle = ipcom().fswatch(self, self.path.parent, False)
            self.count += 1
        return self

    def __exit__(self, *args):
        """Stops monitoring task outputs"""
        with self.lock:
            self.count -= 1
            if self.count == 0:
                ipcom().fsunwatch(self.handle)
                self.fh.close()

                self.handle = None
                self._fh = None

    def watch_output(self, watched: "WatchedOutput"):
        """Add a new listener"""
        key = f"{watched.config.__identifier__}/{watched.method_name}"
        with self.lock:
            # Process events so far
            listener = self.listeners[key].get(watched.method, None)
            if listener is None:
                listener = TaskOutputCallbackHandler(watched.method)

            # Register
            self.listeners[key][watched.method].add(watched.callback)

    #
    # --- Events
    #

    @cached_property
    def fh(self):
        if self._fh is None:
            self._fh = self.path.open("rt")
        return self._fh

    def on_modified(self, event):
        self.handle(Path(event.src_path))

    def on_created(self, event):
        self.handle(Path(event.src_path))

    def handle(self, path: Path):
        if path != self.path:
            return

        with self.lock:
            logger.debug("[TASK OUTPUT] Handling task output for %s", self.path)

            while json_line := self.fh.readline():
                # Read the event
                event = json.loads(json_line)
                logger.debug("Event: %s", event)

                # FIXME: move elsewhere
                # # Process the event
                # event = self.config_method(
                #     self.job.config.__xpm__.mark_output,
                #     *event["args"],
                #     **event["kwargs"],
                # )

                self.events.append(event)
                # self.job.scheduler.xp.taskOutputsWorker.add(self, event)


class TaskOutputsWorker(threading.Thread):
    """This worker process dynamic output queue for one experiment"""

    def __init__(self, xp: experiment):
        super().__init__(name="task outputs worker")
        self.queue = queue.Queue()
        self.xp = xp

    def watch_output(self, watched: "WatchedOutput"):
        """Watch an output

        :param watched: The watched output specification
        """
        logger.debug("Registering task output listener %s", watched)

        # path = watched.job.tasks_output_path
        TaskOutputs.create(watched.job).watch_output(watched)

    def add(self, watcher, event):
        asyncio.run_coroutine_threadsafe(
            self.xp.update_task_output_count(1),
            self.xp.scheduler.loop,
        ).result()
        self.queue.put((watcher, event))

    def run(self):
        logging.debug("Starting output listener queue")
        while True:
            # Get the next element in the queue
            element = self.queue.get()
            if element is None:
                # end of processing
                break

            # Call all the listeners
            logging.debug("Got one event: %s", element)
            watcher, event = element
            for listener in watcher.listeners:
                try:
                    logger.debug("Calling listener [%s] with %s", listener, event)
                    listener(event)
                    logger.debug(
                        "[done] Calling listener [%s] with %s", listener, event
                    )
                except Exception:
                    logging.exception("Exception while calling the listener")
                self.queue.task_done()

                asyncio.run_coroutine_threadsafe(
                    self.xp.update_task_output_count(-1), self.xp.scheduler.loop
                ).result()



import imp
import importlib
import inspect
import json
import logging
import sys
from pathlib import Path
from typing import Any, List, Optional, Protocol, Tuple

import click
import omegaconf
import yaml
from omegaconf import OmegaConf, SCMode
from termcolor import cprint

from experimaestro import LauncherRegistry, RunMode, experiment
from experimaestro.exceptions import HandledException
from experimaestro.experiments.configuration import ConfigurationBase
from experimaestro.settings import find_workspace


class ExperimentHelper:
    """Helper for experiments"""

    xp: experiment
    """The experiment object"""

    #: Run function
    callable: "ExperimentCallable"

    def __init__(self, callable: "ExperimentCallable"):
        self.callable = callable

    """Handles extra arguments"""

    def run(self, args: List[str], configuration: ConfigurationBase):
        assert len(args) == 0
        self.callable(self, configuration)

    @classmethod
    def decorator(cls, *args, **kwargs):
        """Decorator for the run(helper, configuration) method"""
        if len(args) == 1 and len(kwargs) == 0 and inspect.isfunction(args[0]):
            return cls(callable)

        def wrapper(callable):
            return cls(callable)

        return wrapper


class ExperimentCallable(Protocol):
    """Protocol for the run function"""

    def __call__(self, helper: ExperimentHelper, configuration: Any):
        ...


class ConfigurationLoader:
    def __init__(self):
        self.yamls = []
        self.python_path = set()

    def load(self, yaml_file: Path):
        """Loads a YAML file, and parents one if they exist"""
        if not yaml_file.exists() and yaml_file.suffix != ".yaml":
            yaml_file = yaml_file.with_suffix(".yaml")

        with yaml_file.open("rt") as fp:
            _data = yaml.full_load(fp)
        if parent := _data.get("parent", None):
            self.load(yaml_file.parent / parent)

        self.yamls.append(_data)

        for path in _data.get("pythonpath", []):
            path = Path(path)
            if path.is_absolute():
                self.python_path.add(path.resolve())
            else:
                self.python_path.add((yaml_file.parent / path).resolve())


@click.option("--debug", is_flag=True, help="Print debug information")
@click.option("--show", is_flag=True, help="Print configuration and exits")
@click.option(
    "--env",
    help="Define one environment variable",
    type=(str, str),
    multiple=True,
)
@click.option(
    "--host",
    type=str,
    default=None,
    help="Server hostname (default to localhost,"
    " not suitable if your jobs are remote)",
)
@click.option(
    "--run-mode",
    type=click.Choice(RunMode),
    default=RunMode.NORMAL,
    help="Sets the run mode",
)
@click.option(
    "--xpm-config-dir",
    type=Path,
    default=None,
    help="Path for the experimaestro config directory "
    "(if not specified, use $HOME/.config/experimaestro)",
)
@click.option(
    "--port",
    type=int,
    default=None,
    help="Port for monitoring (can be defined in the settings.yaml file)",
)
@click.option(
    "--file", "xp_file", help="The file containing the main experimental code"
)
@click.option(
    "--module-name", "module_name", help="Module containing the experimental code"
)
@click.option(
    "--workspace",
    type=str,
    default=None,
    help="Workspace ID (reads from settings.yaml in experimaestro config)",
)
@click.option(
    "--workdir",
    type=str,
    default=None,
    help="Working environment",
)
@click.option("--conf", "-c", "extra_conf", type=str, multiple=True)
@click.option(
    "--pre-yaml", type=str, multiple=True, help="Add YAML file after the main one"
)
@click.option(
    "--post-yaml", type=str, multiple=True, help="Add YAML file before the main one"
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.argument("yaml_file", metavar="YAML file", type=str)
@click.command()
def experiments_cli(  # noqa: C901
    yaml_file: str,
    xp_file: str,
    host: str,
    port: int,
    xpm_config_dir: Path,
    workdir: Optional[Path],
    workspace: Optional[str],
    env: List[Tuple[str, str]],
    run_mode: RunMode,
    extra_conf: List[str],
    pre_yaml: List[str],
    post_yaml: List[str],
    module_name: Optional[str],
    args: List[str],
    show: bool,
    debug: bool,
):
    """Run an experiment"""

    # --- Set the logger
    logging.getLogger().setLevel(logging.DEBUG if debug else logging.INFO)
    logging.getLogger("xpm.hash").setLevel(logging.INFO)

    # --- Loads the YAML
    conf_loader = ConfigurationLoader()
    for y in pre_yaml:
        conf_loader.load(Path(y))
    conf_loader.load(Path(yaml_file))
    for y in post_yaml:
        conf_loader.load(Path(y))

    # --- Merge the YAMLs
    configuration = OmegaConf.merge(*conf_loader.yamls)
    if extra_conf:
        configuration.merge_with(OmegaConf.from_dotlist(extra_conf))

    # --- Get the XP file
    python_path = list(conf_loader.python_path)
    if module_name is None:
        module_name = configuration.get("module", None)

    if xp_file is None:
        xp_file = configuration.get("file", None)
        if xp_file:
            assert (
                not module_name
            ), "Module name and experiment file are mutually exclusive options"
            xp_file = Path(xp_file)
            if not python_path:
                python_path.append(xp_file.parent)
            logging.info("Using python path: %s", ", ".join(str(s) for s in python_path))

    assert (
        module_name or xp_file
    ), "Either the module name or experiment file should be given"

    # --- Set some options

    if xpm_config_dir is not None:
        assert xpm_config_dir.is_dir()
        LauncherRegistry.set_config_dir(xpm_config_dir)

    # --- Finds the "run" function

    # Modifies the Python path
    for path in python_path:
        sys.path.append(str(path))

    if xp_file:
        if not xp_file.exists() and xp_file.suffix != ".py":
            xp_file = xp_file.with_suffix(".py")
        xp_file: Path = Path(yaml_file).parent / xp_file
        with open(xp_file) as src:
            module_name = xp_file.with_suffix("").name
            mod = imp.load_module(
                module_name,
                src,
                str(xp_file.absolute()),
                (".py", "r", imp.PY_SOURCE),
            )
    else:
        # Module
        try:
            mod = importlib.import_module(module_name)
        except ModuleNotFoundError as e:
            logging.error("Module not found: %s with python path %s", e, sys.path)
            raise

    helper = getattr(mod, "run", None)

    # --- ... and runs it
    if helper is None:
        raise ValueError(f"Could not find run function in {xp_file}")

    if not isinstance(helper, ExperimentHelper):
        helper = ExperimentHelper(helper)

    parameters = inspect.signature(helper.callable).parameters
    list_parameters = list(parameters.values())
    assert len(list_parameters) == 2, (
        "Callable function should only "
        f"have two arguments (got {len(list_parameters)})"
    )

    schema = list_parameters[1].annotation
    omegaconf_schema = OmegaConf.structured(schema())

    if omegaconf_schema is not None:
        try:
            configuration = OmegaConf.merge(omegaconf_schema, configuration)
        except omegaconf.errors.ConfigKeyError as e:
            cprint(f"Error in configuration:\n\n{e}", "red", file=sys.stderr)
            sys.exit(1)

    if show:
        print(json.dumps(OmegaConf.to_container(configuration)))  # noqa: T201
        sys.exit(0)

    # Move to an object container
    configuration = OmegaConf.to_container(
        configuration, structured_config_mode=SCMode.INSTANTIATE
    )

    # Define the workspace
    ws_env = find_workspace(workdir=workdir, workspace=workspace)
    
    workdir = ws_env.path

    logging.info("Using working directory %s", str(workdir.resolve()))
    
    # --- Runs the experiment
    with experiment(
        ws_env, configuration.id, host=host, port=port, run_mode=run_mode
    ) as xp:
        # Set up the environment
        # (1) global settings (2) workspace settings and (3) command line settings
        for key, value in env:
            xp.setenv(key, value)

        # Sets the python path
        xp.workspace.python_path.extend(python_path)

        try:
            # Run the experiment
            helper.xp = xp
            helper.run(list(args), configuration)

            # ... and wait
            xp.wait()

        except HandledException:
            sys.exit(1)



# flake8: noqa: T201
import subprocess
from typing import Optional
from shutil import rmtree
import click
from pathlib import Path
from termcolor import colored, cprint

from experimaestro.settings import find_workspace
from . import check_xp_path, cli


@click.option("--workspace", default="", help="Experimaestro workspace")
@click.option("--workdir", type=Path, default=None)
@cli.group()
@click.pass_context
def jobs(
    ctx,
    workdir: Optional[Path],
    workspace: Optional[str],
):
    """Job control: list, kill and clean

    The job filter is a boolean expression where tags (alphanumeric)
    and special job information (@state for job state, @name for job full
    name) can be compared to a given value (using '~' for regex matching,
    '=', 'not in', or 'in')

    For instance,

    model = "bm25" and mode in ["a", b"] and @state = "RUNNING"

    selects jobs where the tag model is "bm25", the tag mode is either
    "a" or "b", and the state is running.

    """
    ws = ctx.obj.workspace = find_workspace(workdir=workdir, workspace=workspace)
    check_xp_path(ctx, None, ws.path)


def process(
    workspace,
    *,
    experiment="",
    tags="",
    ready=False,
    clean=False,
    kill=False,
    filter="",
    perform=False,
    fullpath=False,
):
    from .filter import createFilter, JobInformation
    from experimaestro.scheduler import JobState

    _filter = createFilter(filter) if filter else lambda x: True

    # Get all jobs from experiments
    job2xp = {}

    path = workspace.path
    for p in (path / "xp").glob("*"):
        for job in p.glob("jobs/*/*"):
            job_path = job.resolve()
            if job_path.is_dir():
                *_, scriptname = job_path.parent.name.rsplit(".", 1)
                job2xp.setdefault(scriptname, set()).add(p.name)

        if (p / "jobs.bak").is_dir():
            cprint(f"  Experiment {p.name} has not finished yet", "red")
            if (not perform) and (kill or clean):
                cprint("  Preventing kill/clean (use --perform if you want to)", "yellow")
                kill = False
                clean = False

    # Now, process jobs
    for job in path.glob("jobs/*/*"):
        info = None
        p = job.resolve()
        if p.is_dir():
            *_, scriptname = p.parent.name.rsplit(".", 1)
            xps = job2xp.get(scriptname, set())
            if experiment and experiment not in xps:
                continue

            info = JobInformation(p, scriptname)
            job_str = (
                (str(job.resolve()) if fullpath else f"{job.parent.name}/{job.name}")
                + " "
                + ",".join(xps)
            )

            if filter:
                if not _filter(info):
                    continue

            if info.state is None:
                print(colored(f"NODIR   {job_str}", "red"), end="")
            elif info.state.running():
                if kill:
                    if perform:
                        process = info.getprocess()
                        if process is None:
                            cprint(
                                "internal error – no process could be retrieved",
                                "red",
                            )
                        else:
                            cprint(f"KILLING {process}", "light_red")
                            process.kill()
                    else:
                        print("KILLING (not performing)", process)
                print(
                    colored(f"{info.state.name:8}{job_str}", "yellow"),
                    end="",
                )
            elif info.state == JobState.DONE:
                print(
                    colored(f"DONE    {job_str}", "green"),
                    end="",
                )
            elif info.state == JobState.ERROR:
                print(colored(f"FAIL    {job_str}", "red"), end="")
            else:
                print(
                    colored(f"{info.state.name:8}{job_str}", "red"),
                    end="",
                )

        else:
            if not ready:
                continue
            print(colored(f"READY {job_path}", "yellow"), end="")

        if tags:
            print(f""" {" ".join(f"{k}={v}" for k, v in info.tags.items())}""")
        else:
            print()

        if clean and info.state and info.state.finished():
            if perform:
                cprint("Cleaning...", "red")
                rmtree(p)
            else:
                cprint("Cleaning... (not performed)", "red")
    print()


@click.option("--experiment", default=None, help="Restrict to this experiment")
@click.option("--tags", is_flag=True, help="Show tags")
@click.option("--ready", is_flag=True, help="Include tasks which are not yet scheduled")
@click.option("--filter", default="", help="Filter expression")
@click.option("--fullpath", is_flag=True, help="Prints full paths")
@jobs.command()
@click.pass_context
def list(
    ctx,
    experiment: str,
    filter: str,
    tags: bool,
    ready: bool,
    fullpath: bool,
):
    process(
        ctx.obj.workspace,
        experiment=experiment,
        filter=filter,
        tags=tags,
        ready=ready,
        fullpath=fullpath,
    )


@click.option("--experiment", default=None, help="Restrict to this experiment")
@click.option("--tags", is_flag=True, help="Show tags")
@click.option("--ready", is_flag=True, help="Include tasks which are not yet scheduled")
@click.option("--filter", default="", help="Filter expression")
@click.option("--perform", is_flag=True, help="Really perform the killing")
@click.option("--fullpath", is_flag=True, help="Prints full paths")
@jobs.command()
@click.pass_context
def kill(
    ctx,
    experiment: str,
    filter: str,
    tags: bool,
    ready: bool,
    fullpath: bool,
    perform: bool,
):
    process(
        ctx.obj.workspace,
        experiment=experiment,
        filter=filter,
        tags=tags,
        ready=ready,
        kill=True,
        perform=perform,
        fullpath=fullpath,
    )


@click.option("--experiment", default=None, help="Restrict to this experiment")
@click.option("--tags", is_flag=True, help="Show tags")
@click.option("--ready", is_flag=True, help="Include tasks which are not yet scheduled")
@click.option("--filter", default="", help="Filter expression")
@click.option("--perform", is_flag=True, help="Really perform the cleaning")
@click.option("--fullpath", is_flag=True, help="Prints full paths")
@jobs.command()
@click.pass_context
def clean(
    ctx,
    experiment: str,
    filter: str,
    tags: bool,
    ready: bool,
    fullpath: bool,
    perform: bool,
):
    process(
        ctx.obj.workspace,
        experiment=experiment,
        filter=filter,
        tags=tags,
        ready=ready,
        clean=True,
        perform=perform,
        fullpath=fullpath,
    )


@click.argument("jobid", type=str)
@click.option(
    "--follow", "-f", help="Use tail instead of less to follow changes", is_flag=True
)
@click.option("--std", help="Follow stdout instead of stderr", is_flag=True)
@jobs.command()
@click.pass_context
def log(ctx, jobid: str, follow: bool, std: bool):
    task_name, task_hash = jobid.split("/")
    _, name = task_name.rsplit(".", 1)
    path = (
        ctx.obj.workspace.path
        / "jobs"
        / task_name
        / task_hash
        / f"""{name}.{'out' if std else 'err'}"""
    )
    if follow:
        subprocess.run(["tail", "-f", path])
    else:
        subprocess.run(["less", "-r", path])


@click.argument("jobid", type=str)
@jobs.command()
@click.pass_context
def path(ctx, jobid: str):
    task_name, task_hash = jobid.split("/")
    path = ctx.obj.workspace.path / "jobs" / task_name / task_hash
    print(path)



# --- Task and types definitions

import logging
from experimaestro import Config, Param
from typing import Union

import pytest
from experimaestro.core.objects import TypeConfig


def test_multiple_inheritance():
    class A(Config):
        pass

    class B(Config):
        pass

    class B1(B):
        pass

    class C1(B1, A):
        pass

    class C2(A, B1):
        pass

    for C in (C1, C2):
        logging.info("Testing %s", C)
        ctype = C.__xpmtype__
        assert issubclass(C, A)
        assert issubclass(C, B)
        assert issubclass(C, B1)

        assert ctype.objecttype == C.__xpmtype__.objecttype

        assert issubclass(C.__xpmtype__.objecttype, B1.__xpmtype__.basetype)
        assert issubclass(C.__xpmtype__.objecttype, B.__xpmtype__.basetype)
        assert issubclass(C.__xpmtype__.objecttype, A.__xpmtype__.basetype)
        assert not issubclass(C.__xpmtype__.objecttype, TypeConfig)


def test_missing_hierarchy():
    class A(Config):
        pass

    class A1(A):
        pass

    class B(A1):
        pass

    B.__xpmtype__

    assert issubclass(B, A)
    assert issubclass(B, A1)


def test_types_union():
    class A(Config):
        x: Param[Union[int, str]]

    A(x=1)
    A(x="hello")
    with pytest.raises(ValueError):
        A(x=[])
